from nmdc_schema.migrators.migrator_base import MigratorBase


class Migrator(MigratorBase):
    r"""
    Migrates a database between two schemas.
    """

    _from_version = "PR104"
    _to_version = "PR195"

    # Map the names of the collections that store documents representing instances of direct
    # subclasses of `PlannedProcess`, to a list of the collections that do/would store documents
    # representing instances of that subclass's subclasses.
    # Reference: https://microbiomedata.github.io/berkeley-schema-fy24/PlannedProcess/
    collection_names_hierarchy = {
        "collecting_biosamples_from_site_set": [],
        "protocol_execution_set": [],
        "storage_process_set": [],
        "material_processing_set": [
            "pooling_set",
            "extraction_set",
            "library_preparation_set",
            "sub_sampling_process_set",
            "mixing_process_set",
            "filtration_process_set",
            "chromatographic_separation_process_set",
            "dissolving_process_set",
            "chemical_conversion_process_set",
        ],
        "data_generation_set": [
            "nucleotide_sequencing_set",
            "mass_spectrometry_set",
        ],
        "workflow_execution_set": [
            "metagenome_annotation_set",
            "metagenome_assembly_set",
            "metatranscriptome_assembly_set",
            "metatranscriptome_annotation_set",
            "metatranscriptome_analysis_set",
            "metatranscriptome_expression_analysis_set",  # included in order to accommodate: https://github.com/microbiomedata/nmdc-schema/pull/2117
            "mags_set",
            "metagenome_sequencing_set",
            "read_qc_analysis_set",
            "read_based_taxonomy_analysis_set",
            "metabolomics_analysis_set",
            "metaproteomics_analysis_set",
            "nom_analysis_set",
        ],
    }

    def upgrade(self) -> None:
        r"""
        Migrates the database from conforming to the original schema, to conforming to the new schema.

        In PR195, the schema changed so that the `Database` class has a slot for each direct subclass of
        `PlannedProcess`; but does not have slots for _their_ subclasses. This migrator creates collections,
        moves documents between collections, and deletes collections, in accordance with that schema change.

        >>> from nmdc_schema.migrators.adapters.dictionary_adapter import DictionaryAdapter
        >>> database = {'material_processing_set': [{'id': 'mp'}],
        ...             'pooling_set': [{'id': 'p'}],
        ...             'metatranscriptome_expression_analysis_set': [{'id': 'ma1'}],
        ...             'nom_analysis_set': [{'id': 'na1'}, {'id': 'na2'}, {'id': 'na3'}]}
        >>> m = Migrator(adapter=DictionaryAdapter(database=database))
        >>> m.upgrade()
        >>> len(database['material_processing_set'])  # ensure migrator moves document to "parent" collection
        2
        >>> 'pooling_set' in database  # ensure migrator deletes "child" collection
        False
        >>> len(database['workflow_execution_set'])  # ensure migrator creates "parent" collection (if nonexistent)
        4
        >>> 'metatranscriptome_expression_analysis_set' in database
        False
        >>> 'nom_analysis_set' in database
        False
        """

        # Iterate over each parent collection name.
        for (
            parent_collection_name,
            child_collection_names,
        ) in self.collection_names_hierarchy.items():
            self.logger.info(f"Processing collection (parent): {parent_collection_name}")

            # Create the parent collection (if it doesn't already exist).
            self.adapter.create_collection(collection_name=parent_collection_name)

            # Iterate over each of this parent's child collection names.
            for child_collection_name in child_collection_names:
                self.logger.info(
                    f"Moving documents from '{child_collection_name}' (child) "
                    f"to '{parent_collection_name}' (parent)"
                )

                # Insert a copy of each document currently in the child collection,
                # into the parent collection.
                self.adapter.do_for_each_document(
                    collection_name=child_collection_name,
                    action=lambda document: self.adapter.insert_document(
                        collection_name=parent_collection_name, document=document
                    ),
                )

                # Delete the child collection, now that the parent collection contains a copy of
                # each of its documents.
                self.logger.info(f"Deleting collection (child): {child_collection_name}")
                self.adapter.delete_collection(collection_name=child_collection_name)



from typing import List, Type

from nmdc_schema.migrators.migrator_base import MigratorBase
from nmdc_schema.migrators.partials.migrator_from_10_2_0_to_11_0_0 import (
    migrator_from_10_2_0_to_11_0_0_part_01,
    migrator_from_10_2_0_to_11_0_0_part_02,
    migrator_from_10_2_0_to_11_0_0_part_03,
    migrator_from_10_2_0_to_11_0_0_part_04,
    migrator_from_10_2_0_to_11_0_0_part_05,
    migrator_from_10_2_0_to_11_0_0_part_06,
    migrator_from_10_2_0_to_11_0_0_part_07,
    migrator_from_10_2_0_to_11_0_0_part_08,
    migrator_from_10_2_0_to_11_0_0_part_09,
    migrator_from_10_2_0_to_11_0_0_part_10,
    migrator_from_10_2_0_to_11_0_0_part_11,
    migrator_from_10_2_0_to_11_0_0_part_12,
    migrator_from_10_2_0_to_11_0_0_part_13,
    migrator_from_10_2_0_to_11_0_0_part_14,
    migrator_from_10_2_0_to_11_0_0_part_15,
    migrator_from_10_2_0_to_11_0_0_part_16,
)

def get_migrator_classes() -> List[Type[MigratorBase]]:
    r"""
    Returns a list of migrator classes in the order in which they (i.e. their `upgrade` methods)
    were designed to be run.

    >>> migrator_classes = get_migrator_classes()
    >>> type(migrator_classes) is list and len(migrator_classes) > 0  # the function returns a list
    True
    >>> from inspect import isclass
    >>> all(isclass(c) for c in migrator_classes)  # each list item is a classes
    True
    >>> all(callable(getattr(c, "upgrade")) for c in migrator_classes)  # each class has an `upgrade` method
    True
    """

    return [
        migrator_from_10_2_0_to_11_0_0_part_01.Migrator,
        migrator_from_10_2_0_to_11_0_0_part_02.Migrator,
        migrator_from_10_2_0_to_11_0_0_part_03.Migrator,
        migrator_from_10_2_0_to_11_0_0_part_04.Migrator,
        migrator_from_10_2_0_to_11_0_0_part_05.Migrator,
        migrator_from_10_2_0_to_11_0_0_part_06.Migrator,
        migrator_from_10_2_0_to_11_0_0_part_07.Migrator,
        migrator_from_10_2_0_to_11_0_0_part_08.Migrator,
        migrator_from_10_2_0_to_11_0_0_part_09.Migrator,
        migrator_from_10_2_0_to_11_0_0_part_10.Migrator,
        migrator_from_10_2_0_to_11_0_0_part_11.Migrator,
        migrator_from_10_2_0_to_11_0_0_part_12.Migrator,
        migrator_from_10_2_0_to_11_0_0_part_13.Migrator,
        migrator_from_10_2_0_to_11_0_0_part_14.Migrator,
        migrator_from_10_2_0_to_11_0_0_part_15.Migrator,
        migrator_from_10_2_0_to_11_0_0_part_16.Migrator
    ]



from typing import Optional

from nmdc_schema.migrators.migrator_base import MigratorBase
from nmdc_schema.nmdc_data import get_nmdc_schema_definition
from linkml_runtime import SchemaView


def create_schema_view() -> SchemaView:
    """
    Returns a LinkML SchemaView instance that can be used to traverse the schema.
    
    >>> isinstance(create_schema_view(), SchemaView)
    True
    """
    schema_definition = get_nmdc_schema_definition()
    schema_view = SchemaView(schema_definition)
    return schema_view


class Migrator(MigratorBase):
    """Migrates data from schema X to PR10"""

    _from_version = "X"
    _to_version = "PR10"

    def upgrade(self):
        r"""
        Migrates the database from conforming to the original schema, to conforming to the new schema.

        >>> from nmdc_schema.migrators.adapters.dictionary_adapter import DictionaryAdapter
        >>> database = {
        ...     "data_object_set": [
        ...         {'id': 1},
        ...         {'id': 2, 'type': 'old'}
        ...     ],
        ...     "library_preparation_set": [
        ...         {'id': 1},
        ...         {'id': 2, 'type': 'old'},
        ...         {'id': 3, 'protocol_link': {'name': 'nombre'}},
        ...         {'id': 4, 'protocol_link': {'name': 'nombre', 'type': 'old'}}
        ...     ],
        ... }
        >>> m = Migrator(adapter=DictionaryAdapter(database=database))
        >>> m.upgrade()
        >>> all(document['type'] == 'nmdc:DataObject' for document in database['data_object_set'])
        True
        >>> all(document['type'] == 'nmdc:LibraryPreparation' for document in database['library_preparation_set'])
        True

        Confirm a `type` field has been added to the inline `protocol_link` instance that lacked one.
        >>> library_prep_3 = next(document for document in database['library_preparation_set'] if document['id'] == 3)
        >>> library_prep_3['protocol_link']['name']
        'nombre'
        >>> library_prep_3['protocol_link']['type']
        'nmdc:Protocol'

        Confirm the `type` value has been updated on the inline `protocol_link` instance that had an incorrect one.
        >>> library_prep_4 = next(document for document in database['library_preparation_set'] if document['id'] == 4)
        >>> library_prep_4['protocol_link']['name']
        'nombre'
        >>> library_prep_4['protocol_link']['type']
        'nmdc:Protocol'
        """

        # Get a dictionary of slots and the class uris of their range if they have inlined classes as their range.
        view = create_schema_view()

        slots_with_inlined_classes = {}
        classes_with_inlined_classes = ["Biosample", 
                                        "Study", 
                                        "Extraction", 
                                        "MetabolomicsAnalysis" , 
                                        "MetaproteomicsAnalysis", 
                                        "MagsAnalysis",
                                        "ReadQcAnalysis",
                                        "DataGeneration",
                                        "LibraryPreparation",
                                        "Pooling",
                                        "MetagenomeAnnotation",
                                        "MetagenomeAssembly",
                                        "MetagenomeSequencing",
                                        "NomAnalysis",
                                        "ReadBasedTaxonomyAnalysis",
                                        "MetatranscriptomeAnnotation",
                                        "MetatranscriptomeAssembly",
                                        "MetatranscriptomeExpressionAnalysis",
                                        "CollectingBiosamplesFromSite"
                                        ]
        for nmdc_class in classes_with_inlined_classes:
            induced_slots = view.class_induced_slots(nmdc_class)
            for slot_def in induced_slots:
                slot_name = slot_def.name
                slot_range = slot_def.range
                element = view.get_element(slot_range)
                metatype = type(element).class_name
                if metatype == 'class_definition':
                    class_identifying_slot = view.get_identifier_slot(slot_range)
                    if not class_identifying_slot:
                        # need to hardcode because schema prefix is not NMDC
                        if slot_name == "has_credit_associations":
                            class_uri = "prov:Association"
                            slots_with_inlined_classes[slot_name] = class_uri
                        else:
                            class_uri = f"{view.schema.default_prefix}:{element.name}"
                            slots_with_inlined_classes[slot_name] = class_uri

        # Populate the "collection-to-transformers" map for this specific migration.
        agenda = dict(
            biosample_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:Biosample", slots_with_inlined_classes)],
            study_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:Study", slots_with_inlined_classes)],
            extraction_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:Extraction", slots_with_inlined_classes)],
            mags_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:MagsAnalysis", slots_with_inlined_classes)],
            metabolomics_analysis_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:MetabolomicsAnalysis", slots_with_inlined_classes)],
            metaproteomics_analysis_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:MetaproteomicsAnalysis", slots_with_inlined_classes)],
            data_generation_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:DataGeneration", slots_with_inlined_classes)],
            read_qc_analysis_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:ReadQcAnalysis", slots_with_inlined_classes)],
            library_preparation_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:LibraryPreparation", slots_with_inlined_classes)],
            pooling_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:Pooling", slots_with_inlined_classes)],
            metagenome_annotation_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:MetagenomeAnnotation", slots_with_inlined_classes)],
            metagenome_assembly_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:MetagenomeAssembly", slots_with_inlined_classes)],
            metagenome_sequencing_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:MetagenomeSequencing", slots_with_inlined_classes)],
            nom_analysis_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:NomAnalysis", slots_with_inlined_classes)],
            read_based_taxonomy_analysis_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:ReadBasedTaxonomyAnalysis", slots_with_inlined_classes)],
            metatranscriptome_annotation_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:MetatranscriptomeAnnotation", slots_with_inlined_classes)],
            metatranscriptome_assembly_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:MetatranscriptomeAssembly", slots_with_inlined_classes)],
            metatranscriptome_expression_analysis_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:MetatranscriptomeExpressionAnalysis", slots_with_inlined_classes)],
            collecting_biosamples_from_site_set=[lambda document: self.add_type_slot_with_class_uri(document, "nmdc:CollectingBiosamplesFromSite", slots_with_inlined_classes)]
        )  
        

        for collection_name, pipeline in agenda.items():
            self.adapter.process_each_document(collection_name=collection_name, pipeline=pipeline)

        # For each collection that is not allowed to contain documents having any slots having inlined classes,
        # we can use this more optimized adapter method to set the `type` field of all documents in that collection.
        self.adapter.set_field_of_each_document("data_object_set", "type", "nmdc:DataObject")
        self.adapter.set_field_of_each_document("functional_annotation_agg", "type", "nmdc:FunctionalAnnotationAggMember")
        self.adapter.set_field_of_each_document("field_research_site_set", "type", "nmdc:FieldResearchSite")
        self.adapter.set_field_of_each_document("processed_sample_set", "type", "nmdc:ProcessedSample")


    def add_type_to_inlined_classes(self, document: dict, slot: str, uri: str):
        r"""
        Adds a type slot to each inlined instance of an NMDC class in the biosample_set, study_set, extraction_set, 
        mags_set, metabolomics_analysis_set, and metaproteomics_analysis_set. 
        """
        if document.get(slot):
            # If slot is a list, iterate over each item in the list (e.g. chem_administration and has_credit_associations)
            if isinstance(document[slot], list):
                for item in document[slot]:
                    # will exclude non-inlined class slots that use an id
                    if not isinstance(item, str):
                        item["type"] = uri
                        if item.get("term"):
                            item["term"]["type"] = "nmdc:OntologyClass"
                        if item.get("applies_to_person"):
                            item["applies_to_person"]["type"] = "nmdc:PersonValue"
            else:
                # If slot is not multivalued
                document[slot]["type"] = uri
                if document[slot].get("term"):
                    document[slot]["term"]["type"] = "nmdc:OntologyClass"
                    

    def add_type_slot_with_class_uri(self, document: dict, class_uri: str, inlined_slots: Optional[dict] = None):
        r"""
            Adds a type slot to each collection with the appropriate class uri as the value. E.g. type: nmdc:Biosample. If a type
            slot exists, it will overwrite to the types listed below. If the optional dictionary argument inlined_slots is given, then
            this function will add the type lost to any slots that have inlined classes as their range.
        
            >>> m = Migrator()
            >>> m.add_type_slot_with_class_uri({'id': 123, 'collection_date': {'has_raw_value': '2017-05-09'}}, 'nmdc:Biosample', {'collection_date': 'nmdc:TimestampValue'}) 
            {'id': 123, 'collection_date': {'has_raw_value': '2017-05-09', 'type': 'nmdc:TimestampValue'}, 'type': 'nmdc:Biosample'}
            >>> m.add_type_slot_with_class_uri({'id': 567, 'type': 'nmdc:DataGeneration'}, 'nmdc:DataGeneration')
            {'id': 567, 'type': 'nmdc:DataGeneration'}
            >>> m.add_type_slot_with_class_uri({'id': 567, 'env_broad_scale': {'term': {'id': 'ENVO:1234'}}}, 'nmdc:Biosample', {'env_broad_scale': 'nmdc:ControlledIdentifiedTermValue'})
            {'id': 567, 'env_broad_scale': {'term': {'id': 'ENVO:1234', 'type': 'nmdc:OntologyClass'}, 'type': 'nmdc:ControlledIdentifiedTermValue'}, 'type': 'nmdc:Biosample'}
            >>> m.add_type_slot_with_class_uri({'id': 456}, 'nmdc:NomAnalysis')
            {'id': 456, 'type': 'nmdc:NomAnalysis'}
        """
        
        # Adds the type slot with the correct class_uri as a value to each collection instance
        document["type"] = class_uri
                
        # Add the type slot to any inlined classes in the Biosample, Study, or Extraction collections
        if inlined_slots:
            for slot, uri in inlined_slots.items():
                self.add_type_to_inlined_classes(document, slot, uri)
                
        return document

    